import argparse
import os
import io

from label_studio_converter.converter import Converter, Format


class ExpandFullPath(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, os.path.abspath(os.path.expanduser(values)))


def main():
    parser = argparse.ArgumentParser(
        description='Converter from Label Studio output completions to various formats',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-i', '--input', dest='input', required=True,
        help='Directory or JSON file with completions (e.g. "/<project_path>/completions")',
        action=ExpandFullPath
    )
    parser.add_argument(
        '-c', '--config', dest='config', required=True,
        help='Project config (e.g. "/<project_path>/config.xml")',
        action=ExpandFullPath
    )
    parser.add_argument(
        '-o', '--output', dest='output',
        help='Output file or directory (will be created if not exists)',
        default=os.path.join(os.path.dirname(__file__), 'output'),
        action=ExpandFullPath
    )
    parser.add_argument(
        '-f', '--format', dest='format',
        metavar='FORMAT',
        help='Output format: ' + ', '.join(f.name for f in Format),
        type=Format.from_string,
        choices=list(Format),
        default=Format.JSON
    )
    parser.add_argument(
        '--csv-separator', dest='csv_separator',
        help='Separator used in CSV format',
        default=','
    )
    parser.add_argument(
        '--csv-no-header', dest='csv_no_header',
        help='Whether to omit header in CSV output file',
        action='store_true'
    )
    parser.add_argument(
        '--image-dir', dest='image_dir',
        help='In case of image outputs (COCO, VOC, ...), specifies output image directory where downloaded images will '
             'be stored. (If not specified, local image paths left untouched)'
    )
    parser.add_argument(
        '--project-dir', dest='project_dir', default=None,
        help='Label Studio project directory path'
    )
    parser.add_argument(
        '--heartex-format', dest='heartex_format', action='store_true',
        help='Set this flag if your completions are coming from Heartex platform instead of Label Studio'
    )
    args = parser.parse_args()

    with io.open(args.config) as f:
        config_str = f.read()
    c = Converter(config_str, project_dir=args.project_dir)

    if args.format == Format.JSON:
        c.convert_to_json(args.input, args.output)
    elif args.format == Format.CSV:
        header = not args.csv_no_header
        sep = args.csv_separator
        c.convert_to_csv(args.input, args.output, sep=sep, header=header, is_dir=not args.heartex_format)
    elif args.format == Format.CONLL2003:
        c.convert_to_conll2003(args.input, args.output, is_dir=not args.heartex_format)
    elif args.format == Format.COCO:
        c.convert_to_coco(args.input, args.output, output_image_dir=args.image_dir, is_dir=not args.heartex_format)
    elif args.format == Format.VOC:
        c.convert_to_voc(args.input, args.output, output_image_dir=args.image_dir, is_dir=not args.heartex_format)
    elif args.format == Format.YOLO:
        c.convert_to_yolo(args.input, args.output, is_dir=not args.heartex_format)

    print('Congratulations! Now check:\n' + args.output)


if __name__ == "__main__":
    main()



import unittest
import label_studio_converter.utils as utils


class TestUtils(unittest.TestCase):
    def test_create_tokens_and_tags_with_eol_tag(self):

        s = 'I need a break\nplease'
        spans = [{'end': 14,
                  'labels': ['Person'],
                  'start': 9,
                  'text': 'break',
                  'type': 'Labels'}]
        tokens, tags = utils.create_tokens_and_tags(s, spans)
        self.assertEqual(tokens[0], "I")
        self.assertEqual(tags[0], "O")
        self.assertEqual(tokens[1], "need")
        self.assertEqual(tags[1], "O")
        self.assertEqual(tokens[2], "a")
        self.assertEqual(tags[2], "O")
        self.assertEqual(tokens[3], "break")
        self.assertEqual(tags[3], "B-Person")
        self.assertEqual(tokens[4], "please")
        self.assertEqual(tags[4], "O")

    def test_create_tokens_and_tags_with_tab_tag(self):

        s = 'I need a tab\tplease'
        spans = [{'end': 12,
                  'labels': ['Person'],
                  'start': 9,
                  'text': 'tab',
                  'type': 'Labels'}]
        tokens, tags = utils.create_tokens_and_tags(s, spans)
        self.assertEqual(tokens[0], "I")
        self.assertEqual(tags[0], "O")
        self.assertEqual(tokens[1], "need")
        self.assertEqual(tags[1], "O")
        self.assertEqual(tokens[2], "a")
        self.assertEqual(tags[2], "O")
        self.assertEqual(tokens[3], "tab")
        self.assertEqual(tags[3], "B-Person")
        self.assertEqual(tokens[4], "please")
        self.assertEqual(tags[4], "O")

    def test_create_tokens_and_tags_for_full_token(self):
        text = 'We gave Jane Smith the ball.'
        spans = [{'end': 12, 'labels': ['Person'], 'start': 9, 'text': 'ane'}]
        tokens, tags = utils.create_tokens_and_tags(text, spans)
        self.assertEqual(tokens, ['We', 'gave', 'Jane', 'Smith', 'the', 'ball', '.'])
        self.assertEqual(tags, ['O', 'O', 'B-Person', 'O', 'O', 'O', 'O'])

    def test_create_tokens_and_tags_with_leading_space(self):
        text = 'We gave Jane Smith the ball.'
        spans = [{'end': 18, 'labels': ['Person'], 'start': 7, 'text': ' Jane Smith'}]
        tokens, tags = utils.create_tokens_and_tags(text, spans)
        self.assertEqual(tokens, ['We', 'gave', 'Jane', 'Smith', 'the', 'ball', '.'])
        self.assertEqual(tags, ['O', 'O', 'B-Person', 'I-Person', 'O', 'O', 'O'])

    def test_create_tokens_and_tags_with_trailing_space(self):
        text = 'We gave Jane Smith the ball.'
        spans = [{'end': 19, 'labels': ['Person'], 'start': 8, 'text': 'Jane Smith '}]
        tokens, tags = utils.create_tokens_and_tags(text, spans)
        self.assertEqual(tokens, ['We', 'gave', 'Jane', 'Smith', 'the', 'ball', '.'])
        self.assertEqual(tags, ['O', 'O', 'B-Person', 'I-Person', 'O', 'O', 'O'])

    def test_create_tokens_and_tags_for_token_with_dollar_sign(self):
        text = 'A meal with $3.99 bill'
        spans = [{'end': 16, 'labels': ['Person'], 'start': 12, 'text': '$3.99'}]
        tokens, tags = utils.create_tokens_and_tags(text, spans)
        self.assertEqual(tokens, ['A', 'meal', 'with', '$', '3.99', 'bill'])
        self.assertEqual(tags, ['O', 'O', 'O', 'B-Person', 'I-Person', 'O'])

    def test_create_tokens_and_tags_for_token_with_slash_sign(self):
        text = 'A meal from Google/Facebook'
        spans = [{'end': 17, 'labels': ['ORG'], 'start': 12, 'text': 'Google'},
                 {'end': 22, 'labels': ['ORG'], 'start': 19, 'text': 'Face'}]
        tokens, tags = utils.create_tokens_and_tags(text, spans)
        self.assertEqual(tokens, ['A', 'meal', 'from', 'Google', '/', 'Facebook'])
        self.assertEqual(tags, ['O', 'O', 'O', 'B-ORG', 'O', 'B-ORG'])


    def test_create_tokens_and_tags_for_token_with_ampersand_sign(self):
        text = "They're the world's most fear;some fighting team,Teenage Mut;ant Ninja Turtles (We're really hip!).They're heroes in the half-shell and they're green,Teenage Mutant Ninja Turtles (Hey, get a grip!),When the evil Shre&dder attacks These Turtle boys don't cut him no slack! Spli\/nter taught them to be ninja teens.. Teen#age Mutant Ninja Turtles (He's a radi@cal rat!), Leonardo leads, Donat#ello does machines, Teenage Mutant Ninja Turt;les (That's a fact, Jack!),Raphael is cool but rude (Gimme a break!), Michel;angelo is a party dude (Party!)"
        spans = [{"start":0,"end":4,"text":"They","labels":["PER"]},{"start":12,"end":17,"text":"world","labels":["PER"]},{"start":57,"end":60,"text":"Mut","labels":["PER"]},{"start":212,"end":216,"text":"Shre","labels":["PER"]},{"start":314,"end":318,"text":"Teen","labels":["PER"]},{"start":272,"end":276,"text":"Spli","labels":["PER"]}]
        tokens, tags = utils.create_tokens_and_tags(text, spans)
        self.assertEqual(tokens[52], 'Shre')
        self.assertEqual(tags[52], 'B-PER')
        self.assertEqual(tags[51], 'O')
        self.assertEqual(tags[53], 'O')


    def test_create_tokens_and_tags_for_token_with_hash_sign(self):
        text = "Mut;ant Ninja Turtles. Teen#age Mutant Ninja Turtles."
        spans = [{"start":0,"end":3,"text":"Mut","labels":["PER"]},
                 {"start":5,"end":7,"text":"ant","labels":["ORG"]},
                 {"start":24,"end":27,"text":"Teen","labels":["PER"]},
                 {"start":29,"end":30,"text":"Shre","labels":["ORG"]}]
        tokens, tags = utils.create_tokens_and_tags(text, spans)
        self.assertEqual(tokens, ['Mut', ';', 'ant', 'Ninja', 'Turtles.', 'Teen', '#', 'age', 'Mutant', 'Ninja', 'Turtles', '.'])
        self.assertEqual(tags, ['B-PER', 'O', 'B-ORG', 'O', 'O', 'B-PER', 'O', 'B-ORG', 'O', 'O', 'O', 'O'])


import os
import io
import logging
import json


from .utils import get_audio_duration, ensure_dir, download, _get_annotator


logger = logging.getLogger(__name__)


def convert_to_asr_json_manifest(input_data, output_dir, data_key, project_dir, upload_dir, download_resources):
    audio_dir_rel = 'audio'
    output_audio_dir = os.path.join(output_dir, audio_dir_rel)
    ensure_dir(output_dir), ensure_dir(output_audio_dir)
    output_file = os.path.join(output_dir, 'manifest.json')
    with io.open(output_file, mode='w') as fout:
        for item in input_data:
            audio_path = item['input'][data_key]
            try:
                audio_path = download(audio_path, output_audio_dir, project_dir=project_dir, upload_dir=upload_dir,
                                      return_relative_path=True, download_resources=download_resources)
                duration = get_audio_duration(os.path.join(output_audio_dir, os.path.basename(audio_path)))
            except:
                logger.info('Unable to download {image_path} or get audio duration. The item {item} will be skipped'.format(
                    image_path=audio_path, item=item
                ), exc_info=True)
                continue

            for texts in iter(item['output'].values()):
                if len(texts) > 0 and 'text' in texts[0]:
                    break

            transcript = texts[0]['text'][0]
            metadata = {
                'audio_filepath': audio_path,
                'duration': duration,
                'text': transcript,
                'annotator': _get_annotator(item, default='')
            }
            json.dump(metadata, fout)
            fout.write('\n')



from label_studio_converter.imports.colors import COLORS


LABELS = """
  <{# TAG_NAME #} name="{# FROM_NAME #}" toName="image">
{# LABELS #}  </{# TAG_NAME #}>
"""

LABELING_CONFIG = """<View>
  <Image name="{# TO_NAME #}" value="$image"/>
{# BODY #}</View>
"""


def generate_label_config(categories, tags, to_name='image', from_name='label', filename=None):
    labels = ''
    for key in sorted(categories.keys()):
        color = COLORS[int(key) % len(COLORS)]
        label = f'    <Label value="{categories[key]}" background="rgba({color[0]}, {color[1]}, {color[2]}, 1)"/>\n'
        labels += label

    body = ''
    for from_name in tags:
        tag_body = str(LABELS) \
            .replace('{# TAG_NAME #}', tags[from_name]) \
            .replace('{# LABELS #}', labels) \
            .replace('{# TO_NAME #}', to_name) \
            .replace('{# FROM_NAME #}', from_name)
        body += f'\n  <Header value="{tags[from_name]}"/>' + tag_body

    config = str(LABELING_CONFIG).replace('{# BODY #}', body).replace('{# TO_NAME #}', to_name)

    if filename:
        with open(filename, 'w') as f:
            f.write(config)

    return config



"""
Original RLE JS code from https://github.com/thi-ng/umbrella/blob/develop/packages/rle-pack/src/index.ts

export const decode = (src: Uint8Array) => {
    const input = new BitInputStream(src);
    const num = input.read(32);
    const wordSize = input.read(5) + 1;
    const rleSizes = [0, 0, 0, 0].map(() => input.read(4) + 1);
    const out = arrayForWordSize(wordSize, num);
    let x, j;
    for (let i = 0; i < num; ) {
        x = input.readBit();
        j = i + 1 + input.read(rleSizes[input.read(2)]);
        if (x) {
            out.fill(input.read(wordSize), i, j);
            i = j;
        } else {
            for (; i < j; i++) {
                out[i] = input.read(wordSize);
            }
        }
    }
    return out;
};

const arrayForWordSize = (ws: number, n: number) => {
    return new (ws < 9 ? Uint8Array : ws < 17 ? Uint16Array : Uint32Array)(n);
};
"""
import os
import uuid
import numpy as np
import logging

from PIL import Image
from collections import defaultdict
from itertools import groupby

logger = logging.getLogger(__name__)


### Brush Export ###


class InputStream:
    def __init__(self, data):
        self.data = data
        self.i = 0

    def read(self, size):
        out = self.data[self.i:self.i + size]
        self.i += size
        return int(out, 2)


def access_bit(data, num):
    """ from bytes array to bits by num position
    """
    base = int(num // 8)
    shift = 7 - int(num % 8)
    return (data[base] & (1 << shift)) >> shift


def bytes2bit(data):
    """ get bit string from bytes data
    """
    return ''.join([str(access_bit(data, i)) for i in range(len(data) * 8)])


def decode_rle(rle, print_params: bool = False):
    """ from LS RLE to numpy uint8 3d image [width, height, channel]
    
    Args:
        print_params (bool, optional): If true, a RLE parameters print statement is suppressed
    """
    input = InputStream(bytes2bit(rle))
    num = input.read(32)
    word_size = input.read(5) + 1
    rle_sizes = [input.read(4) + 1 for _ in range(4)]
    
    if print_params:
        print('RLE params:', num, 'values', word_size, 'word_size', rle_sizes, 'rle_sizes')
        
    i = 0
    out = np.zeros(num, dtype=np.uint8)
    while i < num:
        x = input.read(1)
        j = i + 1 + input.read(rle_sizes[input.read(2)])
        if x:
            val = input.read(word_size)
            out[i:j] = val
            i = j
        else:
            while i < j:
                val = input.read(word_size)
                out[i] = val
                i += 1
    return out


def decode_from_annotation(from_name, results):
    """ from LS annotation to {"tag_name + label_name": [numpy uint8 image (width x height)]}
    """
    layers = {}
    counters = defaultdict(int)
    for result in results:
        key = 'brushlabels' if result['type'].lower() == 'brushlabels' else \
            ('labels' if result['type'].lower() == 'labels' else None)
        if key is None or 'rle' not in result:
            continue

        rle = result['rle']
        width = result['original_width']
        height = result['original_height']
        labels = result[key] if key in result else ['no_label']
        name = from_name + '-' + '-'.join(labels)

        # result count
        i = str(counters[name])
        counters[name] += 1
        name += '-' + i

        image = decode_rle(rle)
        layers[name] = np.reshape(image, [height, width, 4])[:, :, 3]
    return layers


def save_brush_images_from_annotation(task_id, annotation_id, completed_by,
                                      from_name, results, out_dir, out_format='numpy'):
    layers = decode_from_annotation(from_name, results)
    if isinstance(completed_by, dict):
        email = completed_by.get('email', '')
    else:
        email = str(completed_by)
    email = "".join(x for x in email if x.isalnum() or x == '@' or x == '.')  # sanitize filename

    for name in layers:
        filename = os.path.join(out_dir, 'task-' + str(task_id) + '-annotation-' + str(annotation_id)
                                + '-by-' + email + '-' + name)
        image = layers[name]
        logger.debug(f'Save image to {filename}')
        if out_format == 'numpy':
            np.save(filename, image)
        elif out_format == 'png':
            im = Image.fromarray(image)
            im.save(filename + '.png')
        else:
            raise Exception('Unknown output format for brush converter')


def convert_task(item, out_dir, out_format='numpy'):
    """ Task with multiple annotations to brush images, out_format = numpy | png
    """
    for from_name, results in item['output'].items():
        save_brush_images_from_annotation(item['id'], item['annotation_id'], item['completed_by'],
                                          from_name, results, out_dir, out_format)


def convert_task_dir(items, out_dir, out_format='numpy'):
    """ Directory with tasks and annotation to brush images, out_format = numpy | png
    """
    for item in items:
        convert_task(item, out_dir, out_format)


# convert_task_dir('/ls/test/completions', '/ls/test/completions/output', 'numpy')


### Brush Import ###


def bits2byte(arr_str, n=8):
    """ Convert bits back to byte

    :param arr_str:  string with the bit array
    :type arr_str: str
    :param n: number of bits to separate the arr string into
    :type n: int
    :return rle:
    :type rle: list
    """
    rle = []
    numbers = [arr_str[i:i + n] for i in range(0, len(arr_str), n)]
    for i in numbers:
        rle.append(int(i, 2))
    return rle


# Shamelessly plagiarized from https://stackoverflow.com/a/32681075/6051733
def base_rle_encode(inarray):
    """ run length encoding. Partial credit to R rle function.
        Multi datatype arrays catered for including non Numpy
        returns: tuple (runlengths, startpositions, values) """
    ia = np.asarray(inarray)                # force numpy
    n = len(ia)
    if n == 0:
        return None, None, None
    else:
        y = ia[1:] != ia[:-1]               # pairwise unequal (string safe)
        i = np.append(np.where(y), n - 1)   # must include last element posi
        z = np.diff(np.append(-1, i))       # run lengths
        p = np.cumsum(np.append(0, z))[:-1] # positions
        return z, p, ia[i]


def encode_rle(arr, wordsize=8, rle_sizes=[3, 4, 8, 16]):
    """ Encode a 1d array to rle


    :param arr: flattened np.array from a 4d image (R, G, B, alpha)
    :type arr: np.array
    :param wordsize: wordsize bits for decoding, default is 8
    :type wordsize: int
    :param rle_sizes:  list of ints which state how long a series is of the same number
    :type rle_sizes: list
    :return rle: run length encoded array
    :type rle: list

    """
    # Set length of array in 32 bits
    num = len(arr)
    numbits = f'{num:032b}'

    # put in the wordsize in bits
    wordsizebits = f'{wordsize - 1:05b}'

    # put rle sizes in the bits
    rle_bits = ''.join([f'{x - 1:04b}' for x in rle_sizes])

    # combine it into base string
    base_str = numbits + wordsizebits + rle_bits

    # start with creating the rle bite string
    out_str = ''
    for length_reeks, p, value in zip(*base_rle_encode(arr)):
        # TODO: A nice to have but --> this can be optimized but works
        if length_reeks == 1:
            # we state with the first 0 that it has a length of 1
            out_str += '0'
            # We state now the index on the rle sizes
            out_str += '00'

            # the rle size value is 0 for an individual number
            out_str += '000'

            # put the value in a 8 bit string
            out_str += f'{value:08b}'
            state = 'single_val'

        elif length_reeks > 1:
            state = 'series'
            # rle size = 3
            if length_reeks <= 8:
                # Starting with a 1 indicates that we have started a series
                out_str += '1'

                # index in rle size arr
                out_str += '00'

                # length of array to bits
                out_str += f'{length_reeks - 1:03b}'

                out_str += f'{value:08b}'

            # rle size = 4
            elif 8 < length_reeks <= 16:
                # Starting with a 1 indicates that we have started a series
                out_str += '1'
                out_str += '01'

                # length of array to bits
                out_str += f'{length_reeks - 1:04b}'

                out_str += f'{value:08b}'

            # rle size = 8
            elif 16 < length_reeks <= 256:
                # Starting with a 1 indicates that we have started a series
                out_str += '1'

                out_str += '10'

                # length of array to bits
                out_str += f'{length_reeks - 1:08b}'

                out_str += f'{value:08b}'

            # rle size = 16 or longer
            else:

                length_temp = length_reeks
                while length_temp > 2 ** 16:
                    # Starting with a 1 indicates that we have started a series
                    out_str += '1'

                    out_str += '11'
                    out_str += f'{2 ** 16 - 1:016b}'

                    out_str += f'{value:08b}'
                    length_temp -= 2 ** 16

                # Starting with a 1 indicates that we have started a series
                out_str += '1'

                out_str += '11'
                # length of array to bits
                out_str += f'{length_temp - 1:016b}'

                out_str += f'{value:08b}'

    # make sure that we have an 8 fold lenght otherwise add 0's at the end
    nzfill = 8 - len(base_str + out_str) % 8
    total_str = base_str + out_str
    total_str = total_str + nzfill * '0'

    rle = bits2byte(total_str)

    return rle


def contour2rle(contours, contour_id, img_width, img_height):
    """
    :param contours:  list of contours
    :type contours: list
    :param contour_id: id of contour which you want to translate
    :type contour_id: int
    :param img_width: image shape width
    :type img_width: int
    :param img_height: image shape height
    :type img_height: int
    :return: list of ints in RLE format 
    """
    import cv2  # opencv
    mask_im = np.zeros((img_width, img_height, 4))
    mask_contours = cv2.drawContours(mask_im, contours, contour_id, color=(0, 255, 0, 100), thickness=-1)
    rle_out = encode_rle(mask_contours.ravel().astype(int))
    return rle_out


def mask2rle(mask):
    """ Convert mask to RLE
    
    :param mask: uint8 or int np.array mask with len(shape) == 2 like grayscale image
    :return: list of ints in RLE format 
    """
    assert len(mask.shape) == 2, 'mask must be 2D np.array'
    assert mask.dtype == np.uint8 or mask.dtype == int, 'mask must be uint8 or int'
    array = mask.ravel()
    array = np.repeat(array, 4)  # must be 4 channels 
    rle = encode_rle(array)
    return rle
    
    
def image2rle(path):
    """ Convert mask image (jpg, png) to RLE

    1. Read image as grayscale
    2. Flatten to 1d array
    3. Threshold > 128
    4. Encode
    
    :param path: path to image with mask (jpg, png), this image will be thresholded with values > 128 to obtain mask, 
                 so you can mark background as black and foreground as white
    :return: list of ints in RLE format 
    """
    with Image.open(path).convert('L') as image:
        mask = np.array((np.array(image) > 128) * 255, dtype=np.uint8)
        array = mask.ravel()
        array = np.repeat(array, 4)
        rle = encode_rle(array)
        return rle, image.size[0], image.size[1]


def image2annotation(path, label_name, from_name, to_name, ground_truth=False, model_version=None, score=None):
    """ Convert image with mask to brush RLE annotation

    :param path: path to image with mask (jpg, png), this image will be thresholded with values > 128 to obtain mask, 
                 so you can mark background as black and foreground as white
    :param label_name: label name from labeling config (<Label>)
    :param from_name: brush tag name (<BrushLabels>)
    :param to_name: image tag name (<Image>)
    :param ground_truth: ground truth annotation true/false
    :param model_version: any string, only for predictions
    :param score: model score as float, only for predictions
    
    :return: dict with Label Studio Annotation or Prediction (Pre-annotation)
    """
    rle, width, height = image2rle(path)
    result = {
        "result": [
            {
                "id": str(uuid.uuid4())[0:8],
                "type": "brushlabels",
                "value": {
                    "rle": rle,
                    "format": "rle",
                    "brushlabels": [
                        label_name
                    ]
                },
                "origin": "manual",
                "to_name": to_name,
                "from_name": from_name,
                "image_rotation": 0,
                "original_width": width,
                "original_height": height
            }
        ],
    }

    # prediction
    if model_version:
        result["model_version"] = model_version
        result["score"] = score

    # annotation
    else:
        result["ground_truth"] = ground_truth

    return result



""" This code allows to export Label Studio Export JSON to FUNSD format. 
It's only the basic converter, it converts every bbox as a separate word. 

Usage: funsd.py export.json 
This command will export your LS OCR annotations to "./funsd/" directory. 
"""
import os
import json
from collections import defaultdict


def convert_annotation_to_fund(result):
    # collect all LS results and combine labels, text, coordinates into one record
    pre = defaultdict(dict)
    for item in result:
        o = pre[item['id']]

        labels = item.get('value', {}).get('labels', None)
        if labels:
            o['label'] = labels[0]

        text = item.get('value', {}).get('text', None)
        if text:
            o['text'] = text[0]

        if 'box' not in o:
            w, h = item['original_width'], item['original_height']
            v = item.get('value')
            x1 = v['x'] / 100.0 * w
            y1 = v['y'] / 100.0 * h
            x2 = x1 + v['width'] / 100.0 * w
            y2 = y1 + v['height'] / 100.0 * h
            o['box'] = [x1, x2, y1, y2]

    # make FUNSD output
    output = []
    counter = 0
    for key in pre:
        counter += 1
        output.append({
            "id": counter,
            "box": pre[key]['box'],
            "text": pre[key]['text'],
            "label": pre[key]['label'],
            "words": [
                {
                    "box": pre[key]['box'],
                    "text": pre[key]['text']
                }
            ],
            "linking": []
        })

    return {'form': output}


def ls_to_funsd_converter(ls_export_path='export.json', funsd_dir='funsd', data_key='ocr'):
    with open(ls_export_path) as f:
        tasks = json.load(f)

    os.makedirs(funsd_dir, exist_ok=True)

    for task in tasks:
        for annotation in task['annotations']:
            output = convert_annotation_to_fund(annotation['result'])
            filename = task["data"][data_key]
            filename = os.path.basename(filename)
            filename = f'{funsd_dir}/task-{task["id"]}-annotation-{annotation["id"]}-' \
                       f'{filename}.json'

            with open(filename, 'w') as f:
                json.dump(output, f)


if __name__ == '__main__':
	import sys
	print('Usage:', sys.argv[0], 'export.json')
	print('This command will export your LS OCR annotations to "./funsd/" directory')
	
    ls_to_funsd_converter(sys.argv[1])



import io
import os
import requests
import hashlib
import logging
import urllib
import numpy as np
import wave
import shutil
import argparse
import re
import datetime

from operator import itemgetter
from PIL import Image
from urllib.parse import urlparse
from nltk.tokenize.treebank import TreebankWordTokenizer
from lxml import etree
from collections import defaultdict
from label_studio_tools.core.utils.params import get_env

logger = logging.getLogger(__name__)

_LABEL_TAGS = {'Label', 'Choice'}
_NOT_CONTROL_TAGS = {'Filter',}
LOCAL_FILES_DOCUMENT_ROOT = get_env('LOCAL_FILES_DOCUMENT_ROOT', default=os.path.abspath(os.sep))

TreebankWordTokenizer.PUNCTUATION = [
        (re.compile(r"([:,])([^\d])"), r" \1 \2"),
        (re.compile(r"([:,])$"), r" \1 "),
        (re.compile(r"\.\.\."), r" ... "),
        (re.compile(r"[;@#$/%&]"), r" \g<0> "),
        (
            re.compile(r'([^\.])(\.)([\]\)}>"\']*)\s*$'),
            r"\1 \2\3 ",
        ),  # Handles the final period.
        (re.compile(r"[?!]"), r" \g<0> "),
        (re.compile(r"([^'])' "), r"\1 ' "),
    ]

class ExpandFullPath(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, os.path.abspath(os.path.expanduser(values)))


def tokenize(text):
    tok_start = 0
    out = []
    for tok in text.split():
        if len(tok):
            out.append((tok, tok_start))
            tok_start += len(tok) + 1
        else:
            tok_start += 1
    return out


def create_tokens_and_tags(text, spans):
    #tokens_and_idx = tokenize(text) # This function doesn't work properly if text contains multiple whitespaces...
    token_index_tuples = [token for token in TreebankWordTokenizer().span_tokenize(text)]
    tokens_and_idx = [(text[start:end], start) for start, end in token_index_tuples]
    if spans and all([span.get('start') is not None and span.get('end') is not None for span in spans]):
        spans = list(sorted(spans, key=itemgetter('start')))
        span = spans.pop(0)
        span_start = span['start']
        span_end = span['end']-1
        prefix = 'B-'
        tokens, tags = [], []
        for token, token_start in tokens_and_idx:
            tokens.append(token)
            token_end = token_start + len(token) - 1#"- 1" - This substraction is wrong. token already uses the index E.g. "Hello" is 0-4
            token_start_ind = token_start  #It seems like the token start is too early.. for whichever reason

            #if for some reason end of span is missed.. pop the new span (Which is quite probable due to this method)
            #Attention it seems like span['end'] is the index of first char afterwards. In case the whitespace is part of the
            #labell we need to subtract one. Otherwise next token won't trigger the span update.. only the token after next..
            if token_start_ind > span_end:
                while spans:
                    span = spans.pop(0)
                    span_start = span['start']
                    span_end = span['end'] - 1
                    prefix = 'B-'
                    if token_start <= span_end:
                        break
            # Add tag "O" for spans that:
            # - are empty
            # - span start has passed over token_end
            # - do not have any label (None or empty list)
            if not span or token_end < span_start or not span.get('labels'):
                tags.append('O')
            elif span_start <= token_end and span_end >= token_start_ind:
                tags.append(prefix + span['labels'][0])
                prefix = 'I-'
            else:
                tags.append('O')
    else:
        tokens = [token for token, _ in tokens_and_idx]
        tags = ['O'] * len(tokens)

    return tokens, tags


def _get_upload_dir(project_dir=None, upload_dir=None):
    """Return either upload_dir, or path by LS_UPLOAD_DIR, or project_dir/upload"""
    if upload_dir:
        return upload_dir
    upload_dir = os.environ.get('LS_UPLOAD_DIR')
    if not upload_dir and project_dir:
        upload_dir = os.path.join(project_dir, 'upload')
        if not os.path.exists(upload_dir):
            upload_dir = None
    if not upload_dir:
        raise FileNotFoundError("Can't find upload dir: either LS_UPLOAD_DIR or project should be passed to converter")
    return upload_dir


def download(url, output_dir, filename=None, project_dir=None, return_relative_path=False, upload_dir=None,
             download_resources=True):
    is_local_file = url.startswith('/data/') and '?d=' in url
    is_uploaded_file = url.startswith('/data/upload')

    if is_uploaded_file:
        upload_dir = _get_upload_dir(project_dir, upload_dir)
        filename = url.replace('/data/upload/', '')
        filepath = os.path.join(upload_dir, filename)
        logger.debug(f'Copy {filepath} to {output_dir}'.format(filepath=filepath, output_dir=output_dir))
        if download_resources:
            shutil.copy(filepath, output_dir)
        if return_relative_path:
            return os.path.join(os.path.basename(output_dir), filename)
        return filepath

    if is_local_file:
        filename, dir_path = url.split('/data/', 1)[-1].split('?d=')
        dir_path = str(urllib.parse.unquote(dir_path))
        filepath = os.path.join(LOCAL_FILES_DOCUMENT_ROOT, dir_path)
        if not os.path.exists(filepath):
            raise FileNotFoundError(filepath)
        return filepath

    if filename is None:
        basename, ext = os.path.splitext(os.path.basename(urlparse(url).path))
        filename = basename + '_' + hashlib.md5(url.encode() + str(datetime.datetime.now().timestamp()).encode()).hexdigest()[:4] + ext
    filepath = os.path.join(output_dir, filename)
    if not os.path.exists(filepath):
        logger.info('Download {url} to {filepath}'.format(url=url, filepath=filepath))
        if download_resources:
            r = requests.get(url)
            r.raise_for_status()
            with io.open(filepath, mode='wb') as fout:
                fout.write(r.content)
    if return_relative_path:
        return os.path.join(os.path.basename(output_dir), os.path.basename(filename))
    return filepath


def get_image_size(image_path):
    return Image.open(image_path).size


def get_image_size_and_channels(image_path):
    i = Image.open(image_path)
    w, h = i.size
    c = len(i.getbands())
    return w, h, c


def get_audio_duration(audio_path):
    with wave.open(audio_path, mode='r') as f:
        return f.getnframes() / float(f.getframerate())


def ensure_dir(dir_path):
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


def parse_config(config_string):
    """
        :param config_string: Label config string
        :return: structured config of the form:
        {
            "<ControlTag>.name": {
                "type": "ControlTag",
                "to_name": ["<ObjectTag1>.name", "<ObjectTag2>.name"],
                "inputs: [
                    {"type": "ObjectTag1", "value": "<ObjectTag1>.value"},
                    {"type": "ObjectTag2", "value": "<ObjectTag2>.value"}
                ],
                "labels": ["Label1", "Label2", "Label3"] // taken from "alias" if exists or "value"
        }
        """
    if not config_string:
        return {}

    def _is_input_tag(tag):
        return tag.attrib.get('name') and tag.attrib.get('value')

    def _is_output_tag(tag):
        return tag.attrib.get('name') and tag.attrib.get('toName') and tag.tag not in _NOT_CONTROL_TAGS

    def _get_parent_output_tag_name(tag, outputs):
        # Find parental <Choices> tag for nested tags like <Choices><View><View><Choice>...
        parent = tag
        while True:
            parent = parent.getparent()
            if parent is None:
                return
            name = parent.attrib.get('name')
            if name in outputs:
                return name

    try:
        xml_tree = etree.fromstring(config_string)
    except etree.XMLSyntaxError as e:
        raise ValueError(str(e))

    inputs, outputs, labels = {}, {}, defaultdict(dict)
    for tag in xml_tree.iter():
        if _is_output_tag(tag):
            tag_info = {'type': tag.tag, 'to_name': tag.attrib['toName'].split(',')}
            # Grab conditionals if any
            conditionals = {}
            if tag.attrib.get('perRegion') == 'true':
                if tag.attrib.get('whenTagName'):
                    conditionals = {'type': 'tag', 'name': tag.attrib['whenTagName']}
                elif tag.attrib.get('whenLabelValue'):
                    conditionals = {'type': 'label', 'name': tag.attrib['whenLabelValue']}
                elif tag.attrib.get('whenChoiceValue'):
                    conditionals = {'type': 'choice', 'name': tag.attrib['whenChoiceValue']}
            if conditionals:
                tag_info['conditionals'] = conditionals
            outputs[tag.attrib['name']] = tag_info
        elif _is_input_tag(tag):
            inputs[tag.attrib['name']] = {'type': tag.tag, 'value': tag.attrib['value'].lstrip('$')}
        if tag.tag not in _LABEL_TAGS:
            continue
        parent_name = _get_parent_output_tag_name(tag, outputs)
        if parent_name is not None:
            actual_value = tag.attrib.get('alias') or tag.attrib.get('value')
            if not actual_value:
                logger.debug(
                    'Inspecting tag {tag_name}... found no "value" or "alias" attributes.'.format(
                        tag_name=etree.tostring(tag, encoding='unicode').strip()[:50]))
            else:
                labels[parent_name][actual_value] = dict(tag.attrib)
    for output_tag, tag_info in outputs.items():
        tag_info['inputs'] = []
        for input_tag_name in tag_info['to_name']:
            if input_tag_name not in inputs:
                logger.debug(
                    f'to_name={input_tag_name} is specified for output tag name={output_tag}, '
                    'but we can\'t find it among input tags')
                continue
            tag_info['inputs'].append(inputs[input_tag_name])
        tag_info['labels'] = list(labels[output_tag])
        tag_info['labels_attrs'] = labels[output_tag]
    return outputs


def get_polygon_area(x, y):
    """https://en.wikipedia.org/wiki/Shoelace_formula"""

    assert len(x) == len(y)

    return float(0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1))))


def get_polygon_bounding_box(x, y):
    assert len(x) == len(y)

    x1, y1, x2, y2 = min(x), min(y), max(x), max(y)
    return [x1, y1, x2 - x1, y2 - y1]


def _get_annotator(item, default=None, int_id=False):
    """ Get annotator id or email from annotation
    """
    annotator = item['completed_by']
    if isinstance(annotator, dict):
        annotator = annotator.get('email', default)
        return annotator

    if isinstance(annotator, int) and int_id:
        return annotator

    return str(annotator)



import os
import json  # better to use "imports ujson as json" for the best performance
import uuid
import logging
from PIL import Image

from label_studio_converter.utils import ExpandFullPath
from label_studio_converter.imports.label_config import generate_label_config

logger = logging.getLogger('root')


def new_task(out_type, root_url, file_name):
    return {
        "data": {
            "image": os.path.join(root_url, file_name)
        },
        # 'annotations' or 'predictions'
        out_type: [
            {
                "result": [],
                "ground_truth": False,
            }
        ]
    }


def create_bbox(annotation, categories, from_name, image_height, image_width, to_name):
    label = categories[int(annotation['category_id'])]
    x, y, width, height = annotation['bbox']
    x, y, width, height = float(x), float(y), float(width), float(height)
    item = {
        "id": uuid.uuid4().hex[0:10],
        "type": "rectanglelabels",
        "value": {
            "x": x / image_width * 100.0,
            "y": y / image_height * 100.0,
            "width": width / image_width * 100.0,
            "height": height / image_height * 100.0,
            "rotation": 0,
            "rectanglelabels": [
                label
            ]
        },
        "to_name": to_name,
        "from_name": from_name,
        "image_rotation": 0,
        "original_width": image_width,
        "original_height": image_height
    }
    return item


def create_keypoints(annotation, categories, from_name, to_name, image_height, image_width, point_width):
    label = categories[int(annotation['category_id'])]
    points = annotation['keypoints']
    items = []

    for i in range(0, len(points), 3):
        x, y, v = points[i:i+3]  # x, y, visibility
        x, y, v = float(x), float(y), int(v)
        item = {
            "id": uuid.uuid4().hex[0:10],
            "type": "keypointlabels",
            "value": {
                "x": x / image_width * 100.0,
                "y": y / image_height * 100.0,
                "width": point_width,
                "keypointlabels": [
                    label
                ]
            },
            "to_name": to_name,
            "from_name": from_name,
            "image_rotation": 0,
            "original_width": image_width,
            "original_height": image_height
        }

        # visibility
        if v < 2:
            item['value']['hidden'] = True

        items.append(item)
    return items


def convert_coco_to_ls(input_file, out_file,
                       to_name='image', from_name='label', out_type="annotations",
                       image_root_url='/data/local-files/?d=',
                       use_super_categories=False,
                       point_width=1.0):

    """ Convert COCO labeling to Label Studio JSON

    :param input_file: file with COCO json
    :param out_file: output file with Label Studio JSON tasks
    :param to_name: object name from Label Studio labeling config
    :param from_name: control tag name from Label Studio labeling config
    :param out_type: annotation type - "annotations" or "predictions"
    :param image_root_url: root URL path where images will be hosted, e.g.: http://example.com/images
    :param use_super_categories: use super categories from categories if they are presented
    :param point_width: key point width
    """

    tasks = {}  # image_id => task
    logger.info('Reading COCO notes and categories from %s', input_file)

    with open(input_file, encoding='utf8') as f:
        coco = json.load(f)

    # build categories => labels dict
    new_categories = {}
    # list to dict conversion: [...] => {category_id: category_item}
    categories = {int(category['id']): category for category in coco['categories']}
    ids = sorted(categories.keys())  # sort labels by their origin ids

    for i in ids:
        name = categories[i]['name']
        if use_super_categories and 'supercategory' in categories[i]:
            name = categories[i]['supercategory'] + ':' + name
        new_categories[i] = name

    # mapping: id => category name
    categories = new_categories

    # mapping: image id => image
    images = {item['id']: item for item in coco['images']}

    logger.info(f'Found {len(categories)} categories, {len(images)} images and {len(coco["annotations"])} annotations')

    # flags for labeling config composing
    segmentation = bbox = keypoints = rle = False
    segmentation_once = bbox_once = keypoints_once = rle_once = False
    rectangles_from_name, keypoints_from_name = from_name + '_rectangles', from_name + '_keypoints'
    tags = {}

    for i, annotation in enumerate(coco['annotations']):
        segmentation |= 'segmentation' in annotation
        bbox |= 'bbox' in annotation
        keypoints |= 'keypoints' in annotation
        rle |= annotation.get('iscrowd') == 1  # 0 - polygons are in segmentation, otherwise rle

        if rle and not rle_once:  # not supported
            logger.error('RLE in segmentation is not yet supported in COCO')
            rle_once = True
        if keypoints and not keypoints_once:
            logger.warning('Keypoints are partially supported without skeletons')
            tags.update({keypoints_from_name: 'KeyPointLabels'})
            keypoints_once = True
        if segmentation and not segmentation_once:  # not supported
            logger.error('Segmentation is not yet supported in COCO')
            segmentation_once = True
        if bbox and not bbox_once:
            tags.update({rectangles_from_name: 'RectangleLabels'})
            bbox_once = True

        # read image sizes
        image_id = annotation['image_id']
        image = images[image_id]
        image_file_name, image_width, image_height = image['file_name'], image['width'], image['height']

        # get or create new task
        task = tasks[image_id] if image_id in tasks else new_task(out_type, image_root_url, image_file_name)

        if 'bbox' in annotation:
            item = create_bbox(annotation, categories, rectangles_from_name, image_height, image_width, to_name)
            task[out_type][0]['result'].append(item)

        if 'keypoints' in annotation:
            items = create_keypoints(annotation, categories, keypoints_from_name, to_name,
                                     image_height, image_width, point_width)
            task[out_type][0]['result'] += items

        tasks[image_id] = task

    # generate and save labeling config
    label_config_file = out_file.replace('.json', '') + '.label_config.xml'
    generate_label_config(categories, tags, to_name, from_name, label_config_file)

    if len(tasks) > 0:
        tasks = [tasks[key] for key in sorted(tasks.keys())]
        logger.info('Saving Label Studio JSON to %s', out_file)
        with open(out_file, 'w') as out:
            json.dump(tasks, out)

        print('\n'
              f'  1. Create a new project in Label Studio\n'
              f'  2. Use Labeling Config from "{label_config_file}"\n'
              f'  3. Setup serving for images [e.g. you can use Local Storage (or others):\n'
              f'     https://labelstud.io/guide/storage.html#Local-storage]\n'
              f'  4. Import "{out_file}" to the project\n')
    else:
        logger.error('No labels converted')


def add_parser(subparsers):
    coco = subparsers.add_parser('coco')

    coco.add_argument(
        '-i', '--input', dest='input', required=True,
        help='directory with COCO where images, labels, notes.json are located',
        action=ExpandFullPath
    )
    coco.add_argument(
        '-o', '--output', dest='output',
        help='output file with Label Studio JSON tasks',
        default='output.json',
        action=ExpandFullPath
    )
    coco.add_argument(
        '--to-name', dest='to_name',
        help='object name from Label Studio labeling config',
        default='image',
    )
    coco.add_argument(
        '--from-name', dest='from_name',
        help='control tag name from Label Studio labeling config',
        default='label',
    )
    coco.add_argument(
        '--out-type', dest='out_type',
        help='annotation type - "annotations" or "predictions"',
        default='annotations',
    )
    coco.add_argument(
        '--image-root-url', dest='image_root_url',
        help='root URL path where images will be hosted, e.g.: http://example.com/images',
        default='/data/local-files/?d=',
    )
    coco.add_argument(
        '--point-width', dest='point_width',
        help='key point width (size)',
        default=1.0,
        type=float
    )




import setuptools
import label_studio_converter
import os

# Readme
readme_file = os.path.join(os.path.dirname(__file__), 'README.md')
with open(readme_file, 'r') as f:
    long_description = f.read()

# Module dependencies
requirements_file = os.path.join(os.path.dirname(__file__), 'requirements.txt')
with open(requirements_file, 'r') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name='label-studio-converter',
    version=label_studio_converter.__version__,
    author='Heartex',
    author_email="hello@heartex.ai",
    description='Format converter add-on for Label Studio',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/heartexlabs/label-studio-converter',
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    install_requires=requirements,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'label-studio-converter=label_studio_converter.main:main',
        ],
    }
)



import json
import os

from label_studio_converter import Converter
from pandas import read_csv


def test_csv_export():
    # Test case 1, simple output, no JSON
    converter = Converter({}, '/tmp')
    output_dir = '/tmp/lsc-pytest'
    result_csv = output_dir + '/result.csv'
    input_data = os.path.abspath(os.path.dirname(__file__)) + '/data/csv_test.json'
    sep = ','
    converter.convert_to_csv(input_data, output_dir, sep=sep, header=True, is_dir=False)

    df = read_csv(result_csv, sep=sep) 
    nulls = df.isnull().sum()
    if nulls.any() > 0:
        assert False, "There should be no empty values in result CSV"

    # Test case 2, complex fields with JSON
    input_data = os.path.abspath(os.path.dirname(__file__)) + '/data/csv_test2.json'
    sep = '\t'
    converter.convert_to_csv(input_data, output_dir, sep=sep, header=True, is_dir=False)
    df = read_csv(result_csv, sep=sep) 
    nulls = df.isnull().sum()
    assert sum(nulls) == 2, "There should be exactly two empty values in result CSV"

    # Ensure fields are valid JSON
    json.loads(df.iloc[0].writers)
    json.loads(df.iloc[0].iswcs_1)

if __name__ == '__main__':
    test_csv_export()
